# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from unittest import mock

import fixtures
import textwrap


from reno import config
from reno import semver
from reno.tests import base


class TestSemVer(base.TestCase):

    note_bodies = {
        'none': textwrap.dedent("""
        prelude: >
          This should not cause any version update.
        """),
        'major': textwrap.dedent("""
        upgrade:
          - This should cause a major version update.
        """),
        'minor': textwrap.dedent("""
        features:
          - This should cause a minor version update.
        """),
        'patch': textwrap.dedent("""
        fixes:
          - This should cause a patch version update.
        """),
    }

    def _get_note_body(self, filename, sha):
        return self.note_bodies.get(filename, '')

    def _get_dates(self):
        return {'1.0.0': 1547874431}

    def setUp(self):
        super(TestSemVer, self).setUp()
        self.useFixture(
            fixtures.MockPatch('reno.scanner.Scanner.get_file_at_commit',
                               new=self._get_note_body)
        )
        self.useFixture(
            fixtures.MockPatch('reno.scanner.Scanner.get_version_dates',
                               new=self._get_dates)
        )
        self.c = config.Config('.')

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_same(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('1.1.1', []),
        ])
        expected = '1.1.1'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_same_with_note(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('1.1.1', [('none', 'shaA')]),
        ])
        expected = '1.1.1'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_major_working_copy(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('major', 'shaA')]),
            ('1.1.1', []),
        ])
        expected = '2.0.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_major_working_and_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('none', 'shaA')]),
            ('1.1.1-1', [('major', 'shaA')]),
        ])
        expected = '2.0.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_major_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('1.1.1-1', [('major', 'shaA')]),
        ])
        expected = '2.0.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_minor_working_copy(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('minor', 'shaA')]),
            ('1.1.1', []),
        ])
        expected = '1.2.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_minor_working_and_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('none', 'shaA')]),
            ('1.1.1-1', [('minor', 'shaA')]),
        ])
        expected = '1.2.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_minor_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('1.1.1-1', [('minor', 'shaA')]),
        ])
        expected = '1.2.0'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_patch_working_copy(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('patch', 'shaA')]),
            ('1.1.1', []),
        ])
        expected = '1.1.2'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_patch_working_and_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('*working-copy*', [('none', 'shaA')]),
            ('1.1.1-1', [('patch', 'shaA')]),
        ])
        expected = '1.1.2'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    def test_patch_post_release(self, mock_get_notes):
        mock_get_notes.return_value = collections.OrderedDict([
            ('1.1.1-1', [('patch', 'shaA')]),
        ])
        expected = '1.1.2'
        actual = semver.compute_next_version(self.c)
        self.assertEqual(expected, actual)



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import argparse
import os
from unittest import mock

import fixtures
from testtools import ExpectedException

from reno import config
from reno.config import Section
from reno import defaults
from reno import main
from reno.tests import base


def expected_options(**overrides):
    """The default config options, along with any overrides set via kwargs."""
    result = {
        o.name: (
            Section.from_raw_yaml(o.default)
            if o.name == "sections"
            else o.default
        )
        for o in config._OPTIONS
    }
    result.update(**overrides)
    return result


class TestConfig(base.TestCase):
    EXAMPLE_CONFIG = """
collapse_pre_releases: false
"""

    def setUp(self):
        super(TestConfig, self).setUp()
        # Temporary directory to store our config
        self.tempdir = self.useFixture(fixtures.TempDir())

    def test_defaults(self):
        c = config.Config(self.tempdir.path)
        actual = c.options
        self.assertEqual(expected_options(), actual)

    def test_override(self):
        c = config.Config(self.tempdir.path)
        c.override(
            collapse_pre_releases=False,
        )
        actual = c.options
        expected = expected_options(collapse_pre_releases=False)
        self.assertEqual(expected, actual)

    def test_override_multiple(self):
        c = config.Config(self.tempdir.path)
        c.override(
            notesdir='value1',
        )
        c.override(
            notesdir='value2',
        )
        actual = c.options
        expected = expected_options(notesdir='value2')
        self.assertEqual(expected, actual)

    def test_override_sections_with_subsections(self):
        c = config.Config(self.tempdir.path)
        c.override(
            sections=[
                ["features", "Features"],
                ["features_sub", "Sub", 2],
                ["features_subsub", "Subsub", 3],
                ["bugs", "Bugs"],
                ["bugs_sub", "Sub", 2],
                ["documentation", "Documentation", 1]
            ],
        )
        actual = c.options
        expected = expected_options(
            sections=[
                Section("features", "Features", section_level=1),
                Section("features_sub", "Sub", section_level=2),
                Section("features_subsub", "Subsub", section_level=3),
                Section("bugs", "Bugs", section_level=1),
                Section("bugs_sub", "Sub", section_level=2),
                Section("documentation", "Documentation", section_level=1),
            ]
        )
        self.assertEqual(expected, actual)

        # Also check data validation.
        with ExpectedException(ValueError):
            c.override(sections=[["features"]])
        with ExpectedException(ValueError):
            c.override(sections=[["features", "Features", 0]])
        with ExpectedException(ValueError):
            c.override(sections=[["features", "Features", 5]])

    def test_load_file_not_present(self):
        missing = 'reno.config.Config._report_missing_config_files'
        with mock.patch(missing) as error_handler:
            config.Config(self.tempdir.path)
            self.assertEqual(1, error_handler.call_count)

    def _test_load_file(self, config_path):
        with open(config_path, 'w') as fd:
            fd.write(self.EXAMPLE_CONFIG)
        self.addCleanup(os.unlink, config_path)
        c = config.Config(self.tempdir.path)
        self.assertEqual(False, c.collapse_pre_releases)

    def test_load_file_in_releasenotesdir(self):
        rn_path = self.tempdir.join('releasenotes')
        os.mkdir(rn_path)
        config_path = self.tempdir.join('releasenotes/config.yaml')
        self._test_load_file(config_path)

    def test_load_file_in_repodir(self):
        config_path = self.tempdir.join('reno.yaml')
        self._test_load_file(config_path)

    def test_load_file_empty(self):
        config_path = self.tempdir.join('reno.yaml')
        with open(config_path, 'w') as fd:
            fd.write('# Add reno config here')
        self.addCleanup(os.unlink, config_path)
        c = config.Config(self.tempdir.path)
        self.assertEqual(True, c.collapse_pre_releases)

    def test_get_default(self):
        d = config.Config.get_default('notesdir')
        self.assertEqual('notes', d)

    def test_get_default_unknown(self):
        self.assertRaises(
            ValueError,
            config.Config.get_default,
            'unknownopt',
        )

    def _run_override_from_parsed_args(self, argv):
        parser = argparse.ArgumentParser()
        main._build_query_arg_group(parser)
        args = parser.parse_args(argv)
        c = config.Config(self.tempdir.path)
        c.override_from_parsed_args(args)
        return c

    def test_override_from_parsed_args_empty(self):
        c = self._run_override_from_parsed_args([])
        actual = {
            o.name: getattr(c, o.name)
            for o in config._OPTIONS
        }
        self.assertEqual(expected_options(), actual)

    def test_override_from_parsed_args_boolean_false(self):
        c = self._run_override_from_parsed_args([
            '--no-collapse-pre-releases',
        ])
        actual = c.options
        expected = expected_options(collapse_pre_releases=False)
        self.assertEqual(expected, actual)

    def test_override_from_parsed_args_boolean_true(self):
        c = self._run_override_from_parsed_args([
            '--collapse-pre-releases',
        ])
        actual = c.options
        expected = expected_options(collapse_pre_releases=True)
        self.assertEqual(expected, actual)

    def test_override_from_parsed_args_string(self):
        c = self._run_override_from_parsed_args([
            '--earliest-version', '1.2.3',
        ])
        actual = c.options
        expected = expected_options(earliest_version='1.2.3')
        self.assertEqual(expected, actual)

    def test_override_from_parsed_args_ignore_non_options(self):
        parser = argparse.ArgumentParser()
        main._build_query_arg_group(parser)
        parser.add_argument('not_a_config_option')
        args = parser.parse_args(['value'])
        c = config.Config(self.tempdir.path)
        c.override_from_parsed_args(args)
        self.assertFalse(hasattr(c, 'not_a_config_option'))


class TestConfigProperties(base.TestCase):

    def setUp(self):
        super(TestConfigProperties, self).setUp()
        # Temporary directory to store our config
        self.tempdir = self.useFixture(fixtures.TempDir())
        self.c = config.Config('releasenotes')

    def test_reporoot(self):
        self.c.reporoot = 'blah//'
        self.assertEqual('blah/', self.c.reporoot)
        self.c.reporoot = 'blah'
        self.assertEqual('blah/', self.c.reporoot)

    def test_notespath(self):
        self.assertEqual('releasenotes/notes', self.c.notespath)
        self.c.override(notesdir='thenotes')
        self.assertEqual('releasenotes/thenotes', self.c.notespath)

    def test_template(self):
        template = defaults.TEMPLATE.format(defaults.PRELUDE_SECTION_NAME)
        self.assertEqual(template, self.c.template)
        self.c.override(template='i-am-a-template')
        self.assertEqual('i-am-a-template', self.c.template)

    def test_prelude_override(self):
        template = defaults.TEMPLATE.format(defaults.PRELUDE_SECTION_NAME)
        self.assertEqual(template, self.c.template)
        self.c.override(prelude_section_name='fake_prelude_name')
        expected_template = defaults.TEMPLATE.format('fake_prelude_name')
        self.assertEqual(expected_template, self.c.template)

    def test_prelude_and_template_override(self):
        template = defaults.TEMPLATE.format(defaults.PRELUDE_SECTION_NAME)
        self.assertEqual(template, self.c.template)
        self.c.override(prelude_section_name='fake_prelude_name',
                        template='i-am-a-template')
        self.assertEqual('fake_prelude_name', self.c.prelude_section_name)
        self.assertEqual('i-am-a-template', self.c.template)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os.path
import textwrap
from typing import Any
from typing import List
from typing import NamedTuple
from typing import Union

import yaml

from reno import defaults

LOG = logging.getLogger(__name__)


class Opt(NamedTuple):
    name: str
    default: Any
    help: str


class Section(NamedTuple):
    name: str
    title: str
    section_level: int  # 1 represents top-level; higher values are subsctions

    @classmethod
    def from_raw_yaml(
        cls, val: List[List[Union[str, int]]]
    ) -> List["Section"]:
        result = []
        for entry in val:
            if len(entry) == 2:
                section_level = 1
            elif len(entry) == 3:
                section_level = entry[2]
                if (
                    not isinstance(section_level, int)
                    or section_level < 1
                    or section_level > 3
                ):
                    raise ValueError(
                        "The third argument for each entry in the `sections` "
                        "option config must be an integer between 1 and 3."
                        f"Invalid entry: {entry}"
                    )
            else:
                raise ValueError(
                    "Each entry in the `sections` option config must be a "
                    f"list with 2 or 3 values. Invalid entry: {entry}"
                )
            result.append(
                Section(
                    name=entry[0], title=entry[1], section_level=section_level
                )
            )
        return result

    def header_underline(self) -> str:
        symbol = {
            1: "-",
            2: "^",
            3: "~",
        }[self.section_level]
        return symbol * len(self.title)


_OPTIONS = [
    Opt('notesdir', defaults.NOTES_SUBDIR,
        textwrap.dedent("""\
        The notes subdirectory within the relnotesdir where the
        notes live.
        """)),

    Opt('allow_subdirectories', False,
        textwrap.dedent("""\
        Allow creating subdirectories under the notes subdirectory.
        """)),

    Opt('collapse_pre_releases', True,
        textwrap.dedent("""\
        Should pre-release versions be merged into the final release
        of the same number (1.0.0.0a1 notes appear under 1.0.0).
        """)),

    Opt('stop_at_branch_base', True,
        textwrap.dedent("""\
        Should the scanner stop at the base of a branch (True) or go
        ahead and scan the entire history (False)?
        """)),

    Opt('branch', None,
        textwrap.dedent("""\
        The git branch to scan. Defaults to the "current" branch
        checked out. If a stable branch is specified but does not
        exist, reno attempts to automatically convert that to an
        "end-of-life" tag. For example, ``origin/stable/liberty``
        would be converted to ``liberty-eol``.
        """)),

    Opt('default_branch', 'master',
        textwrap.dedent("""\
        The default git branch for the repository. This is the base branch that
        is treated as the root for other branches. By default this is
        ``master``.
        """)),

    Opt('earliest_version', None,
        textwrap.dedent("""\
        The earliest version to be included. This is usually the
        lowest version number, and is meant to be the oldest
        version. If unset, all versions will be scanned.
        """)),

    Opt('template', defaults.TEMPLATE.format(defaults.PRELUDE_SECTION_NAME),
        textwrap.dedent("""\
        The template used by reno new to create a note.
        """)),

    Opt('add_release_date', False,
        textwrap.dedent("""\
        Should the report include release date (True) based on
        the date of objects associated with the release tag.
        """)),

    Opt('release_tag_re',
        textwrap.dedent('''\
        ((?:v?[\\d.ab]|rc)+)  # digits, a, b, and rc cover regular and
                           # pre-releases
        '''),
        textwrap.dedent("""\
        The regex pattern used to match the repo tags representing a
        valid release version. The pattern is compiled with the
        verbose and unicode flags enabled.
        """)),

    Opt('pre_release_tag_re',
        textwrap.dedent('''\
        (?P<pre_release>\\.v?\\d+(?:[ab]|rc)+\\d*)$
        '''),
        textwrap.dedent("""\
        The regex pattern used to check if a valid release version tag
        is also a valid pre-release version. The pattern is compiled
        with the verbose and unicode flags enabled. The pattern must
        define a group called 'pre_release' that matches the
        pre-release part of the tag and any separator, e.g for
        pre-release version '12.0.0.0rc1' the default pattern will
        identify '.0rc1' as the value of the group 'pre_release'.
        """)),

    Opt('branch_name_re', 'stable/.+',
        textwrap.dedent("""\
        The pattern for names for branches that are relevant when
        scanning history to determine where to stop, to find the
        "base" of a branch. Other branches are ignored.
        """)),

    Opt('closed_branch_tag_re', '(.+)-eo[lm]',
        textwrap.dedent("""\
        The pattern for names for tags that replace closed
        branches that are relevant when scanning history to
        determine where to stop, to find the "base" of a
        branch. Other tags are ignored.
        """)),

    Opt('branch_name_prefix', 'stable/',
        textwrap.dedent("""\
        The prefix to add to tags for closed branches
        to restore the old branch name to allow sorting
        to place the tag in the proper place in history.
        For example, OpenStack turns "mitaka-eol" into
        "stable/mitaka" by removing the "-eol" suffix
        via closed_branch_tag_re and setting the prefix
        to "stable/".
        """)),

    Opt('branch_sort_re', 'stable/([0-9].*)',
        textwrap.dedent("""\
        By default branches are sorted alphabetically, except
        for branches matching this pattern, those will be sorted
        with branch_sort_prefix inserted in order to accomodate
        the way OpenStack stable branches are named and sorted.
        """)),

    Opt('branch_sort_prefix', 'stable/zzz',
        textwrap.dedent("""\
        The prefix to add to names of branches matched
        by branch_sort_re. This allows OpenStack branches
        to be sorted according to the current release
        naming scheme. Set to "stable/" in order to
        restore plain alphabetic ordering.
        """)),

    Opt('sections',
        [
            ['features', 'New Features'],
            ['issues', 'Known Issues'],
            ['upgrade', 'Upgrade Notes'],
            ['deprecations', 'Deprecation Notes'],
            ['critical', 'Critical Issues'],
            ['security', 'Security Issues'],
            ['fixes', 'Bug Fixes'],
            ['other', 'Other Notes'],
        ],
        textwrap.dedent("""\
        The identifiers and names of permitted sections in the
        release notes, in the order in which the final report will
        be generated. A prelude section will always be automatically
        inserted before the first element of this list.

        You can optionally include a number from 1 to 3 at the
        end of the list to mark the entry as a subsection. By default,
        each section has the number 1, which represents a top-level
        section. Use 2 and 3 for subsections and subsubsections.
        For example, ``['features', 'New Features', 1]``,
        ``['features_command_line', 'Command Line', 2]``,
        and ``['features_command_line_ios', 'iOS', 3]``. The order of this
        option matters; define subsections right after their ancestor
        sections.

        Warning: you should check that ``semver_major``, ``semver_minor``,
        and ``semver_patch`` includes the relevant section names,
        including subsections.
        """)),

    Opt('prelude_section_name', defaults.PRELUDE_SECTION_NAME,
        textwrap.dedent("""\
        The name of the prelude section in the note template. This
        allows users to rename the section to, for example,
        'release_summary' or 'project_wide_general_announcements',
        which is displayed in titlecase in the report after
        replacing underscores with spaces.
        """)),

    Opt('ignore_null_merges', True,
        textwrap.dedent("""\
        When this option is set to True, any merge commits with no
        changes and in which the second or later parent is tagged
        are considered "null-merges" that bring the tag information
        into the current branch but nothing else.

        OpenStack used to use null-merges to bring final release
        tags from stable branches back into the master branch. This
        confuses the regular traversal because it makes that stable
        branch appear to be part of master and/or the later stable
        branch. This option allows us to ignore those.
        """)),

    Opt('ignore_notes', [],
        textwrap.dedent("""\
        Note files to be ignored. It's useful to be able to ignore a
        file if it is edited on the wrong branch. Notes should be
        specified by their filename or UID.

        Setting the option in the main configuration file makes it
        apply to all branches. To ignore a note in the HTML build, use
        the ``ignore-notes`` parameter to the ``release-notes`` sphinx
        directive.
        """)),

    Opt('unreleased_version_title', '',
        textwrap.dedent("""\
        The title to use for any notes that do not appear in a
        released version. If this option is unset, the development
        version number is used (for example, ``3.0.0-3``).
        """)),
    Opt('encoding', None,
        textwrap.dedent("""\
        The character encoding to use when opening note files. If not
        specified it will be dependent on the system running reno (whatever
        'locale.getpreferredencoding()' returns. This takes in a string
        name that will be passed to the encoding kwarg for open(), so any
        codec or alias from stdlib's codec module is valid.
        """)),

    Opt('semver_major', ['upgrade'],
        textwrap.dedent("""\
        The sections that indicate release notes triggering major version
        updates for the next release, from X.Y.Z to X+1.0.0.
        """)),
    Opt('semver_minor', ['features'],
        textwrap.dedent("""\
        The sections that indicate release notes triggering minor version
        updates for the next release, from X.Y.Z to X.Y+1.0.
        """)),
    Opt('semver_patch', ['fixes'],
        textwrap.dedent("""\
        The sections that indicate release notes triggering patch version
        updates for the next release, from X.Y.Z to X.Y.Z+1.
        """)),
]


class Config:

    _OPTS = {o.name: o for o in _OPTIONS}

    @classmethod
    def get_default(cls, opt):
        """Return the default for an option."""
        try:
            return cls._OPTS[opt].default
        except KeyError:
            raise ValueError('unknown option name %r' % (opt,))

    def __init__(self, reporoot, relnotesdir=None):
        """Instantiate a Config object

        :param str reporoot:
            The root directory of the repository.
        :param str relnotesdir:
            The directory containing release notes. Defaults to
            'releasenotes'.
        """
        self.reporoot = reporoot
        if relnotesdir is None:
            relnotesdir = defaults.RELEASE_NOTES_SUBDIR
        self.relnotesdir = relnotesdir
        # Initialize attributes from the defaults.
        self.override(**{o.name: o.default for o in _OPTIONS})

        self._contents = {}
        self._load_file()

    def _load_file(self):
        filenames = [
            os.path.join(self.reporoot, self.relnotesdir, 'config.yaml'),
            os.path.join(self.reporoot, 'reno.yaml')]

        for filename in filenames:
            LOG.debug('looking for configuration file %s', filename)
            if os.path.isfile(filename):
                break
        else:
            self._report_missing_config_files(filenames)
            return

        try:
            with open(filename, 'r') as fd:
                self._contents = yaml.safe_load(fd)
            LOG.info('loaded configuration file %s', filename)
        except IOError as err:
            self._report_failure_config_file(filename, err)
        else:
            if self._contents:
                self.override(**self._contents)

    def _report_missing_config_files(self, filenames):
        # NOTE(dhellmann): This is extracted so we can mock it for
        # testing.
        LOG.info('no configuration file in: %s', ', '.join(filenames))

    def _report_failure_config_file(self, filename, err):
        # NOTE(dhellmann): This is extracted so we can mock it for
        # testing.
        LOG.warning('did not load config file %s: %s', filename, err)

    def _rename_prelude_section(self, **kwargs):
        key = 'prelude_section_name'
        if key in kwargs and kwargs[key] != self._OPTS[key].default:
            new_prelude_name = kwargs[key]

            self.template = defaults.TEMPLATE.format(new_prelude_name)

    def override(self, **kwds):
        """Set the values of the named configuration options.

        Take the values of the keyword arguments as the current value
        of the same option, regardless of whether a value is already
        present.

        """
        # Replace prelude section name if it has been changed.
        self._rename_prelude_section(**kwds)

        for name, val in kwds.items():
            if name not in self._OPTS:
                LOG.warning('ignoring unknown configuration value %r = %r',
                            name, val)
            else:
                if name == "sections":
                    val = Section.from_raw_yaml(val)
                setattr(self, name, val)

    def override_from_parsed_args(self, parsed_args):
        """Set the values of the configuration options from parsed CLI args.

        This method assumes that the DEST values for the command line
        arguments are named the same as the configuration options.

        """
        arg_values = {
            o.name: getattr(parsed_args, o.name)
            for o in _OPTIONS
            if getattr(parsed_args, o.name, None) is not None
        }
        if arg_values:
            LOG.info('[config] updating from command line options')
            self.override(**arg_values)

    @property
    def reporoot(self):
        return self._reporoot

    # Ensure that the 'reporoot' value always only ends in one '/'.
    @reporoot.setter
    def reporoot(self, value):
        self._reporoot = value.rstrip('/') + '/'

    @property
    def notespath(self):
        """The path in the repo where notes are kept.

        .. important::

           This does not take ``reporoot`` into account. You need to add this
           manually if required.
        """
        return os.path.join(self.relnotesdir, self.notesdir)

    @property
    def options(self):
        """Get all configuration options as a dict.

        Returns the actual configuration options after overrides.
        """
        options = {
            o.name: getattr(self, o.name)
            for o in _OPTIONS
        }
        return options

# def parse_config_into(parsed_arguments):

#         """Parse the user config onto the namespace arguments.

#     :param parsed_arguments:
#         The result of calling :meth:`argparse.ArgumentParser.parse_args`.
#     :type parsed_arguments:
#         argparse.Namespace
#     """
#     config_path = get_config_path(parsed_arguments.relnotesdir)
#     config_values = read_config(config_path)

#     for key in config_values.keys():
#         try:
#             getattr(parsed_arguments, key)
#         except AttributeError:
#             LOG.info('Option "%s" does not apply to this particular command.'
#                      '. Ignoring...', key)
#             continue
#         setattr(parsed_arguments, key, config_values[key])

#     parsed_arguments._config = config_values



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import textwrap
from unittest import mock

import fixtures
import yaml

from reno import config
from reno import loader
from reno.tests import base


class TestValidate(base.TestCase):

    scanner_output = {
        '0.0.0': [('note', 'shaA')],
    }

    versions = ['0.0.0']

    def setUp(self):
        super(TestValidate, self).setUp()
        self.logger = self.useFixture(
            fixtures.FakeLogger(
                format='%(message)s',
                level=logging.WARNING,
            )
        )
        self.c = config.Config('reporoot')

    def _make_loader(self, note_bodies):
        def _load(ldr):
            ldr._scanner_output = self.scanner_output
            ldr._cache = {
                'file-contents': {'note1': note_bodies},
            }

        with mock.patch('reno.loader.Loader._load_data', _load):
            return loader.Loader(
                self.c,
                ignore_cache=False,
            )

    def test_note_with_non_prelude_string_converted_to_list(self):
        """Test behavior when a non-prelude note is not structured as a list.

        We should silently convert it to list.
        """
        note_bodies = yaml.safe_load(textwrap.dedent("""
        issues: |
          This is a single string. It should be converted to a list.
        """))
        self.assertIsInstance(note_bodies['issues'], str)
        with self._make_loader(note_bodies) as ldr:
            parse_results = ldr.parse_note_file('note1', None)
        self.assertIsInstance(parse_results['issues'], list)

    def test_invalid_note_with_prelude_as_list(self):
        note_bodies = yaml.safe_load(textwrap.dedent('''
        prelude:
          - The prelude should not be a list.
        '''))
        self.assertIsInstance(note_bodies['prelude'], list)
        with self._make_loader(note_bodies) as ldr:
            ldr.parse_note_file('note1', None)
        self.assertIn('does not parse as a single string', self.logger.output)

    def test_invalid_note_with_colon_as_dict(self):
        note_bodies = yaml.safe_load(textwrap.dedent('''
        issues:
          - This line is fine.
          - dict: But this is parsed as a mapping (dictionary), which is bad.
        '''))
        self.assertIsInstance(note_bodies['issues'][-1], dict)
        with self._make_loader(note_bodies) as ldr:
            ldr.parse_note_file('note1', None)
        self.assertIn('instead of a string', self.logger.output)

    def test_invalid_note_with_unrecognized_key(self):
        """Test behavior when note contains an unrecognized section."""
        note_bodies = yaml.safe_load(textwrap.dedent('''
        foobar:
        - |
          This is an issue but we're using an unrecognized section key.
        '''))
        self.assertIsInstance(note_bodies, dict)
        with self._make_loader(note_bodies) as ldr:
            ldr.parse_note_file('note1', None)
        self.assertIn(
            'The foobar section of note1 is not a recognized section.',
            self.logger.output)

    def test_invalid_note_with_missing_key(self):
        """Test behavior when note is not structured as a mapping.

        This one should be an error since we can't correct the input.
        """
        note_bodies = yaml.safe_load(textwrap.dedent('''
        - |
          This is an issue but we're missing the top-level 'issues' key.
        '''))
        self.assertIsInstance(note_bodies, list)
        with self._make_loader(note_bodies) as ldr:
            self.assertRaises(ValueError, ldr.parse_note_file, 'note1', None)
        self.assertIn(
            'does not appear to be structured as a YAML mapping',
            self.logger.output)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from datetime import datetime
import logging
import os.path

import yaml

from reno import scanner

LOG = logging.getLogger(__name__)


def get_cache_filename(conf):
    return os.path.normpath(os.path.join(
        conf.reporoot, conf.notespath, 'reno.cache'))


class Loader(object):
    "Load the release notes for a given repository."

    def __init__(self, conf, ignore_cache=False):
        """Initialize a Loader.

        The versions are presented in reverse chronological order.

        Notes files are associated with the earliest version for which
        they were available, regardless of whether they changed later.

        :param conf: Parsed configuration from file
        :type conf: reno.config.Config
        :param ignore_cache: Do not load a cache file if it is present.
        :type ignore_cache: bool
        """
        self._config = conf
        self._ignore_cache = ignore_cache

        self._reporoot = conf.reporoot
        self._notespath = conf.notespath
        self._branch = conf.branch
        self._collapse_pre_releases = conf.collapse_pre_releases
        self._earliest_version = conf.earliest_version

        self._cache = None
        self._scanner = None
        self._scanner_output = None
        self._tags_to_dates = None
        self._cache_filename = get_cache_filename(conf)
        self._encoding = conf.options['encoding']

        self._load_data()

    def _load_data(self):
        cache_file_exists = os.path.exists(self._cache_filename)

        if self._ignore_cache and cache_file_exists:
            LOG.debug('ignoring cache file %s', self._cache_filename)

        if (not self._ignore_cache) and cache_file_exists:
            LOG.debug('loading cache file %s', self._cache_filename)

            with open(self._cache_filename, 'r', encoding=self._encoding) as f:
                self._cache = yaml.safe_load(f.read())

        if self._cache:
            # Save the cached scanner output to the same attribute
            # it would be in if we had loaded it "live". This
            # simplifies some of the logic in the other methods.
            self._scanner_output = collections.OrderedDict(
                (n['version'], n['files'])
                for n in self._cache['notes']
            )
            self._tags_to_dates = collections.OrderedDict(
                (n['version'], n['date'])
                for n in self._cache['dates']
            )
        else:
            self._scanner = scanner.Scanner(self._config)
            self._scanner_output = self._scanner.get_notes_by_version()
            self._tags_to_dates = self._scanner.get_version_dates()

    def close(self):
        """Close any files opened by this loader."""
        if self._scanner is not None:
            self._scanner.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    @property
    def versions(self):
        "A list of all of the versions found."
        return list(self._scanner_output.keys())

    def __getitem__(self, version):
        "Return data about the files that should go into a given version."
        return self._scanner_output[version]

    def get_version_date(self, version):
        "Return release data for a version."
        if version in self._tags_to_dates.keys():
            date = datetime.fromtimestamp(self._tags_to_dates[version])
            return date.strftime("%Y-%m-%d")
        return "Unknown"

    def parse_note_file(self, filename, sha):
        """Return the data structure encoded in the note file.

        Emit warnings for content that does not look valid in some
        way, but return it anyway for backwards-compatibility.

        """
        if self._cache:
            content = self._cache['file-contents'][filename]
        else:
            body = self._scanner.get_file_at_commit(filename, sha)
            content = yaml.safe_load(body)

        cleaned_content = {}

        if not isinstance(content, dict):
            LOG.warning(
                '%s does not appear to be structured as a YAML mapping. '
                'Did you forget a top-level key?',
                filename,
            )
            raise ValueError(
                f'{filename} does not appear to be structured as a YAML '
                f'mapping. Did you forget a top-level key?'
            )

        valid_section_names = {
            section.name for section in self._config.sections
        }
        for section_name, section_content in content.items():
            if section_name == self._config.prelude_section_name:
                if not isinstance(section_content, str):
                    LOG.warning(
                        'The %s section of %s does not parse as a single '
                        'string. Is the YAML input escaped properly?',
                        section_name, filename,
                    )
            else:
                if section_name not in valid_section_names:
                    # TODO(stephenfin): Make this an error in a future release
                    LOG.warning(
                        'The %s section of %s is not a recognized section. '
                        'It should be one of: %s. '
                        'This will be an error in a future release.',
                        section_name, filename,
                        ', '.join(valid_section_names),
                    )
                if isinstance(section_content, str):
                    # A single string is OK, but wrap it with a list
                    # so the rest of the code can treat the data model
                    # consistently.
                    section_content = [section_content]
                elif not isinstance(section_content, list):
                    LOG.warning(
                        'The %s section of %s does not parse as a string or '
                        'list of strings. Is the YAML input escaped properly?',
                        section_name, filename,
                    )
                else:
                    for item in section_content:
                        if not isinstance(item, str):
                            LOG.warning(
                                'The item %r in the %s section of %s parses '
                                'as a %s instead of a string. '
                                'Is the YAML input escaped properly?',
                                item, section_name, filename, type(item),
                            )

            cleaned_content[section_name] = section_content

        return cleaned_content



# -*- coding: utf-8 -*-

# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import collections
from unittest import mock

import fixtures
import textwrap


from reno import cache
from reno import config
from reno.tests import base


class TestCache(base.TestCase):

    scanner_output = [
        collections.OrderedDict([  # master
            ('0.0.0', [('note1', 'shaA')]),
            ('1.0.0', [('note2', 'shaB'), ('note3', 'shaC')]),
        ]),
        collections.OrderedDict([  # stable/1.0
            ('1.0.1', [('note4', 'shaD')]),
        ]),
    ]

    note_bodies = {
        'note1': textwrap.dedent("""
        prelude: >
          This is the prelude.
        """),
        'note2': textwrap.dedent("""
        issues:
          - This is the first issue.
          - This is the second issue.
        """),
        'note3': textwrap.dedent("""
        features:
          - We added a feature!
        """),
        'note4': textwrap.dedent("""
        fixes:
          - We fixed all the bugs!
        """),
    }

    def _get_note_body(self, filename, sha):
        return self.note_bodies.get(filename, '')

    def _get_dates(self):
        return {'1.0.0': 1547874431}

    def setUp(self):
        super(TestCache, self).setUp()
        self.useFixture(
            fixtures.MockPatch('reno.scanner.Scanner.get_file_at_commit',
                               new=self._get_note_body)
        )
        self.useFixture(
            fixtures.MockPatch('reno.scanner.Scanner.get_version_dates',
                               new=self._get_dates)
        )
        self.c = config.Config('.')

    @mock.patch('reno.scanner.Scanner.get_notes_by_version')
    @mock.patch('reno.scanner.Scanner.get_series_branches')
    def test_build_cache_db(self, mock_get_branches, mock_get_notes):
        mock_get_notes.side_effect = self.scanner_output
        mock_get_branches.return_value = ['stable/1.0']
        expected = {
            'dates': [{'version': '1.0.0', 'date': 1547874431}],
            'notes': [
                {'version': '0.0.0',
                 'files': [('note1', 'shaA')]},
                {'version': '1.0.0',
                 'files': [('note2', 'shaB'), ('note3', 'shaC')]},
                {'version': '1.0.1',
                 'files': [('note4', 'shaD')]},
            ],
            'file-contents': {
                'note1': {
                    'prelude': 'This is the prelude.\n',
                },
                'note2': {
                    'issues': [
                        'This is the first issue.',
                        'This is the second issue.',
                    ],
                },
                'note3': {
                    'features': ['We added a feature!'],
                },
                'note4': {
                    'fixes': ['We fixed all the bugs!'],
                },
            },
        }

        db = cache.build_cache_db(
            self.c,
            versions_to_include=[],
        )

        mock_get_branches.assert_called_once()
        mock_get_notes.assert_has_calls([
            mock.call(None), mock.call('stable/1.0')])
        self.assertEqual(expected, db)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from packaging import version

from reno import loader

LOG = logging.getLogger(__name__)


def compute_next_version(conf):
    "Compute the next semantic version based on the available release notes."
    LOG.debug('starting semver-next')
    with loader.Loader(conf, ignore_cache=True) as ldr:
        LOG.debug('known versions: %s', ldr.versions)

        # We want to include any notes in the local working directory or
        # in any commits that came after the last tag. We should never end
        # up with more than 2 entries in to_include.
        to_include = []
        for to_consider in ldr.versions:
            if to_consider == '*working-copy*':
                to_include.append(to_consider)
                continue

            # This check relies on PEP 440 versioning
            parsed = version.Version(to_consider)
            if parsed.post:
                to_include.append(to_consider)
                continue

            break

        # If we found no commits then we're sitting on a real tag and
        # there is nothing to do to update the version.
        if not to_include:
            LOG.debug('found no staged notes and no post-release commits')
            return ldr.versions[0]

        LOG.debug('including notes from %s', to_include)

        candidate_bases = to_include[:]
        if candidate_bases[0] == '*working-copy*':
            candidate_bases = candidate_bases[1:]

        if not candidate_bases:
            # We have a real tag and some locally modified files. Use the
            # real tag as the basis of the next version.
            base_version = version.Version(ldr.versions[1])
        else:
            base_version = version.Version(candidate_bases[0])

        LOG.debug('base version %s', base_version)

        inc_minor = False
        inc_patch = False

        for ver in to_include:
            for filename, sha in ldr[ver]:
                notes = ldr.parse_note_file(filename, sha)

                for section in conf.semver_major:
                    if notes.get(section, []):
                        LOG.debug('found breaking change in %r section of %s',
                                  section, filename)
                        return '{}.0.0'.format(base_version.major + 1)

                for section in conf.semver_minor:
                    if notes.get(section, []):
                        LOG.debug('found feature in %r section of %s',
                                  section, filename)
                        inc_minor = True
                        break

                for section in conf.semver_patch:
                    if notes.get(section, []):
                        LOG.debug('found bugfix in %r section of %s',
                                  section, filename)
                        inc_patch = True
                        break

    major = base_version.major
    minor = base_version.minor
    patch = base_version.micro

    if inc_patch:
        patch += 1

    if inc_minor:
        minor += 1
        patch = 0

    return '{}.{}.{}'.format(major, minor, patch)


def semver_next_cmd(args, conf):
    "Calculate next semantic version number"
    print(compute_next_version(conf))
    return 0



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import os.path

from docutils import nodes
from docutils.parsers import rst
from docutils.parsers.rst import directives
from docutils import statemachine
from dulwich import repo
from sphinx.util import logging
from sphinx.util.nodes import nested_parse_with_titles

import reno
from reno import config
from reno import defaults
from reno import formatter
from reno import loader

LOG = logging.getLogger(__name__)


class ReleaseNotesDirective(rst.Directive):

    has_content = True

    # FIXME(dhellmann): We should be able to build this information
    # from the configuration options so we don't have to edit it
    # manually when we add new options.
    option_spec = {
        'branch': directives.unchanged,
        'reporoot': directives.unchanged,
        'relnotessubdir': directives.unchanged,
        'notesdir': directives.unchanged,
        'version': directives.unchanged,
        'collapse-pre-releases': directives.flag,
        'earliest-version': directives.unchanged,
        'stop-at-branch-base': directives.flag,
        'ignore-notes': directives.unchanged,
        'unreleased-version-title': directives.unchanged,
    }

    def _find_reporoot(self, reporoot_opt, relnotessubdir_opt):
        """Find root directory of project."""
        reporoot = os.path.abspath(reporoot_opt)
        # When building on RTD.org the root directory may not be
        # the current directory, so look for it.
        try:
            return repo.Repo.discover(reporoot).path
        except Exception:
            pass

        for root in ('.', '..', '../..'):
            if os.path.exists(os.path.join(root, relnotessubdir_opt)):
                return root

        raise Exception(
            'Could not discover root directory; tried: %s' % ', '.join([
                os.path.abspath(root) for root in ('.', '..', '../..')
            ])
        )

    def run(self):
        title = ' '.join(self.content)
        branch = self.options.get('branch')
        relnotessubdir = self.options.get(
            'relnotessubdir', defaults.RELEASE_NOTES_SUBDIR,
        )
        reporoot = self._find_reporoot(
            self.options.get('reporoot', '.'), relnotessubdir,
        )
        ignore_notes = [
            name.strip()
            for name in self.options.get('ignore-notes', '').split(',')
        ]
        conf = config.Config(reporoot, relnotessubdir)
        opt_overrides = {}
        if 'notesdir' in self.options:
            opt_overrides['notesdir'] = self.options.get('notesdir')
        version_opt = self.options.get('version')
        # FIXME(dhellmann): Force these flags True for now and figure
        # out how Sphinx passes a "false" flag later.
        # 'collapse-pre-releases' in self.options
        opt_overrides['collapse_pre_releases'] = True
        # Only stop at the branch base if we have not been told
        # explicitly which versions to include.
        opt_overrides['stop_at_branch_base'] = (version_opt is None)
        if 'earliest-version' in self.options:
            opt_overrides['earliest_version'] = self.options.get(
                'earliest-version')
        if 'unreleased-version-title' in self.options:
            opt_overrides['unreleased_version_title'] = self.options.get(
                'unreleased-version-title')

        if branch:
            opt_overrides['branch'] = branch
        if ignore_notes:
            opt_overrides['ignore_notes'] = ignore_notes
        conf.override(**opt_overrides)

        notesdir = os.path.join(relnotessubdir, conf.notesdir)
        LOG.info('scanning %s for %s release notes' % (
                 os.path.join(conf.reporoot, notesdir),
                 branch or 'current branch'))

        with loader.Loader(conf) as ldr:
            if version_opt is not None:
                versions = [
                    v.strip()
                    for v in version_opt.split(',')
                ]
            else:
                versions = ldr.versions
            LOG.info('got versions %s' % (versions,))
            text = formatter.format_report(
                ldr,
                conf,
                versions,
                title=title,
                branch=branch,
            )

        source_name = '<%s %s>' % (__name__, branch or 'current branch')
        result = statemachine.ViewList()
        for line_num, line in enumerate(text.splitlines(), 1):
            LOG.debug('%4d: %s', line_num, line)
            result.append(line, source_name, line_num)

        node = nodes.section()
        node.document = self.state.document
        nested_parse_with_titles(self.state, result, node)
        return node.children


def setup(app):
    app.add_directive('release-notes', ReleaseNotesDirective)
    metadata_dict = {
        'version': reno.__version__,
        'parallel_read_safe': True
    }
    return metadata_dict



# -*- coding: utf-8 -*-
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = [
    'openstackdocstheme',
    'sphinx.ext.autodoc',
    'reno.sphinxext',
    'reno._exts.show_reno_config',
]

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'reno'
copyright = '2013, OpenStack Foundation'

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# Do not warn about non-local image URI
suppress_warnings = ['image.nonlocal_uri']


# -- openstackdocstheme configuration -----------------------------------------

html_theme = 'openstackdocs'

openstackdocs_repo_name = 'openstack/reno'
openstackdocs_use_storyboard = True



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging

from reno import loader

LOG = logging.getLogger(__name__)


def list_cmd(args, conf):
    "List notes files based on query arguments"
    LOG.debug('starting list')
    reporoot = conf.reporoot
    with loader.Loader(conf) as ldr:
        if args.version:
            versions = args.version
        else:
            versions = ldr.versions
        for version in versions:
            notefiles = ldr[version]
            print(version)
            for n, sha in notefiles:
                if n.startswith(reporoot):
                    n = n[len(reporoot):]
                print('\t%s (%s)' % (n, sha))
    return
